/*

  P-Code interpreter (to run the apple pascal system)
  Copyright (C) 2000 Mario Klebsch

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <assert.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "psystem.h"
#include "Memory.h"
#include "Diskio.h"

static int DskTable[16]={0,14,13,12,11,10,9,8,7,6,5,4,3,2,1,15};

static struct
{
  byte		*Data;
  size_t	Size;
  int		RdOnly;
  int		*Translate;
} Units[MAX_UNIT];

void DiskRead(word Unit, word Addr, Integer AddrOffset,
	      word Len, word BlockNo, word Mode)
{
  int		i;
  size_t	Offset = BlockNo*512;
  int		Track  = BlockNo/8;
  int		Sector = (BlockNo&7)*2;

  assert(Unit<MAX_UNIT);
  if (!Units[Unit].Data)
    {
      IoError(9);
      return;
    }
  if ((Offset+Len)>Units[Unit].Size)
    {
      IoError(64);
      return;
    }
  while (Len)
    {
      int Size=256;
      int Sec=Sector;
      if (Len<Size)
	Size=Len;
      if (Units[Unit].Translate)
	Sec=Units[Unit].Translate[Sector];
      for (i=0; i<Size; i++)
	MemWrByte(Addr, AddrOffset+i, Units[Unit].Data[(Track*16+Sec)*256+i]);
      AddrOffset +=Size;
      Len-=Size;
      Sector++;
      if (Sector>=16)
	{
	  Track++;
	  Sector=0;
	}
    }
  IoError(0);
}

void DiskWrite(word Unit, word Addr, Integer AddrOffset,
	       word Len, word BlockNo, word Mode)
{
  int		i;
  size_t	Offset = BlockNo*512;
  int		Track  = BlockNo/8;
  int		Sector = (BlockNo&7)*2;

  assert(Unit<MAX_UNIT);
  if (!Units[Unit].Data)
    IoError(9);
  else if (Units[Unit].RdOnly)
    IoError(16);
  else if ((Offset+Len)>Units[Unit].Size)
    IoError(64);
  else
    {
      while (Len)
	{
	  int Size=256;
	  int Sec=Sector;
	  if (Len<Size)
	    Size=Len;
	  if (Units[Unit].Translate)
	    Sec=Units[Unit].Translate[Sector];
	  for (i=0; i<Size; i++)
	    Units[Unit].Data[(Track*16+Sec)*256+i] =
	      MemRdByte(Addr, AddrOffset+i );
	  AddrOffset +=Size;
	  Len-=Size;
	  Sector++;
	  if (Sector>=16)
	    {
	      Track++;
	      Sector=0;
	    }
	}
      IoError(0);
    }
}

void DiskClear(word Unit)
{
  assert(Unit<MAX_UNIT);
  IoError(0);
}

void DiskStat(word Unit)
{
  assert(Unit<MAX_UNIT);
  if (Units[Unit].Data)
    IoError(0);
  else
    IoError(9);
}

void DiskUmount(word Unit)
{
  assert(Unit<MAX_UNIT);
  if (Units[Unit].Data)
    {
      munmap(Units[Unit].Data, Units[Unit].Size);
      Units[Unit].Data=NULL;
    }
}

int DiskMount(word Unit, const char *FileName, enum DiskMode Mode)
{
  int	fd;
  struct stat buf;

  assert(Unit<MAX_UNIT);
  if ((fd=open(FileName,(Mode==ReadWrite)?O_RDWR:O_RDONLY))<0)
    {
      perror(FileName);
      return(-1);
    }
  if (fstat(fd, &buf)<0)
    {
      perror(FileName);
      close(fd);
      return(-1);
    }
  DiskUmount(Unit);

  Units[Unit].Size=buf.st_size;
  Units[Unit].Data=mmap(NULL, Units[Unit].Size,
			(Mode==ReadOnly)?PROT_READ:PROT_READ|PROT_WRITE,
			(Mode==Forget)?MAP_PRIVATE:MAP_SHARED,
			fd, 0);
  Units[Unit].RdOnly=(Mode==ReadOnly);
  Units[Unit].Translate=NULL;

  if ( (strlen(FileName)>4) &&
       (strcmp(FileName+strlen(FileName)-4,".dsk")==0) )
    Units[Unit].Translate=DskTable;
  close(fd);
  return(0);
}

void DiskInit(void)
{
  int i;
  for (i=0; i<MAX_UNIT; i++)
    Units[i].Data=NULL;
}
