/*

  P-Code interpreter (to run the apple pascal system)
  Copyright (C) 2000 Mario Klebsch

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "psystem.h"
#include "Memory.h"
#include "Stack.h"

#define XSIZE 280
#define YSIZE 192

#define PENCOLOR_NONE		0
#define PENCOLOR_WHITE		1
#define PENCOLOR_BLACK		2
#define PENCOLOR_REVERSE	3
#define PENCOLOR_RADAR		4
#define PENCOLOR_BLACK1		5
#define PENCOLOR_GREEN		6
#define PENCOLOR_VIOLET		7
#define PENCOLOR_WHITE1		8
#define PENCOLOR_BLACK2		9
#define PENCOLOR_ORANGE		10
#define PENCOLOR_BLUE		11
#define PENCOLOR_WHITE2		12

#define NUMBER(a)	(sizeof(a)/sizeof(a[0]))

static const char	*ColorNames[]={NULL, "white", "black", NULL,  NULL, 
				       "black", "green",  "violet", "white",
				       "black", "orange", "blue",   "white"};

static GC		ColorGCs[NUMBER(ColorNames)];

static struct
{
  Integer	xMin;
  Integer	xMax;
  Integer	yMin;
  Integer	yMax;
} CurrentViewPort;
static word	CurrentPenColor;
static Integer	CurrentTurtleX;
static Integer	CurrentTurtleY;
static Integer	CurrentTurtleAng;
static Integer	CurrentCharType;

static unsigned char	*Buffer  = NULL;
static Display		*display = NULL;
static Window		window   = 0;

static void UpdatePixel(int x, int y)
{
  int Pixel=Buffer[y*280/2+x/2];
  if (x&1)
    Pixel >>= 4;
  else
    Pixel &= 0x0f;

  if (Pixel<NUMBER(ColorGCs) &&
      ColorGCs[Pixel])
    XFillRectangle(display, window, ColorGCs[Pixel], x*2, (191-y)*2, 2, 2);
}

static void UpdatePlane(int xmin, int xmax, int ymin, int ymax)
{
  int x, y;

  for (y=ymin; y<ymax; y++)
    for (x=xmin; x<xmax; x++)
      UpdatePixel(x, y);
}

static int GetPixel(int x, int y)
{
  char ch=Buffer[y*280/2+x/2];

  if ((x&1))
    return ((ch&0xf0)>>4);
  else
    return (ch&0x0f);
}

static void SetPixel(int x, int y, int Color)
{
  char			*p=&Buffer[y*280/2+x/2];
  static const char	Inverse[NUMBER(ColorNames)] =
				{0, 2, 1, 0, 0, 8, 7, 6, 5, 12, 11, 10, 9};

  if ( (x<CurrentViewPort.xMin) || (x>CurrentViewPort.xMax) ||
       (y<CurrentViewPort.yMin) || (y>CurrentViewPort.yMax) )
    return;

  if (Color == PENCOLOR_NONE)
    return;
  else if (Color == PENCOLOR_REVERSE)
    Color=Inverse[GetPixel(x,y)];

  if ((x&1))
    *p=(*p&0x0f)|(Color<<4);
  else
    *p=(*p&0xf0)|Color;
  UpdatePixel(x, y);
}

static void Line(int len, int ticks, int *LenVec, int *TickVec)
{
  int	i=len/2;
  int	j;

  for (j=0;j<len; j++)
    {
      CurrentTurtleX+=LenVec[0];
      CurrentTurtleY+=LenVec[1];
      i+=ticks;
      if (i>=len)
	{
	  i-=len;
	  CurrentTurtleX+=TickVec[0];
	  CurrentTurtleY+=TickVec[1];
	}
      SetPixel(CurrentTurtleX, CurrentTurtleY, CurrentPenColor);
    }
}

static void MoveTo(int x, int y)
{
  static int right[] ={ 1,  0};
  static int left[]  ={-1,  0};
  static int up[]    ={ 0,  1};
  static int down[]  ={ 0, -1};

  if (display && (CurrentPenColor!=PENCOLOR_NONE) )
    {
      int dx=x-CurrentTurtleX;
      int dy=y-CurrentTurtleY;
      if ( (dx>0) && (abs(dy)<=dx) )
	Line(dx, abs(dy), right, (dy>0)?up:down);
      else if ( (dx<0) && (abs(dy)<=-dx) )
	Line(-dx, abs(dy), left, (dy>0)?up:down);
      else if ( (dy>0) && (abs(dx<=dy)) )
	Line(dy, abs(dx), up, (dx>0)?right:left);
      else
	Line(-dy, abs(dx), down, (dx>0)?right:left);
    }
  CurrentTurtleX=x;
  CurrentTurtleY=y;
}

void TurnTo(int Angle)
{
  CurrentTurtleAng=Angle;
  while (CurrentTurtleAng<0)
    CurrentTurtleAng+=360;
  while (CurrentTurtleAng>=360)
    CurrentTurtleAng-=360;
}

static void FillScreen(word Color)
{
  int	x, y;
  
  for (y=CurrentViewPort.yMin; y<=CurrentViewPort.yMax; y++)
    for (x=CurrentViewPort.xMin; x<=CurrentViewPort.xMax; x++)
      SetPixel(x, y, Color);
}

static void DrawByte(char Data, int XSkip, int Width,
		     int XScreen, int YScreen, int Mode)
{
  static const int	ModeTable[16][2]={
    {PENCOLOR_BLACK, PENCOLOR_BLACK},	{PENCOLOR_REVERSE, PENCOLOR_BLACK},
    {PENCOLOR_BLACK, PENCOLOR_REVERSE}, {PENCOLOR_REVERSE, PENCOLOR_REVERSE},
    {PENCOLOR_NONE,  PENCOLOR_BLACK},   {PENCOLOR_WHITE,   PENCOLOR_BLACK},
    {PENCOLOR_NONE,  PENCOLOR_REVERSE}, {PENCOLOR_WHITE,   PENCOLOR_REVERSE},
    {PENCOLOR_BLACK, PENCOLOR_NONE},    {PENCOLOR_REVERSE, PENCOLOR_NONE},
    {PENCOLOR_BLACK, PENCOLOR_WHITE},	{PENCOLOR_REVERSE, PENCOLOR_WHITE},
    {PENCOLOR_NONE,  PENCOLOR_NONE},	{PENCOLOR_WHITE,   PENCOLOR_NONE},
    {PENCOLOR_NONE,  PENCOLOR_WHITE},	{PENCOLOR_WHITE,   PENCOLOR_WHITE}   };
  int i;

  for (i=0;i<Width; i++)
    {
      int Pixel;
      if ( (i+XSkip) > 7 )
	break;
      Pixel = Data & (1 << (i+XSkip) ) ? 1:0;
      SetPixel(XScreen+i, YScreen, ModeTable[Mode][Pixel]);
    }
}

static void WChar(unsigned char ch)
{
  int	y;
  static const char	SystemCharset[256*8]={
#include "system.charset.h"
  };

  for (y=0; y<8; y++)
    DrawByte(SystemCharset[ch*8+y], 0, 7,
	     CurrentTurtleX, CurrentTurtleY+y, CurrentCharType);
  CurrentTurtleX += 7;
}

static void InitTurtle(int Width, int Height)
{
  int			screen;
  XSetWindowAttributes	set_win_attr;
  long			win_attr_mask;
  XSizeHints		the_size_hints;
  XGCValues		GcValues;
  int			i;

  if (!display)
    {
      if (!(Buffer=(char*)malloc(280*192/2)))
	{
	  fprintf(stderr,"malloc() failed\n");
	  exit(1);
	}
      memset(Buffer, PENCOLOR_BLACK*0x11, 280*192/2);

      if ( !(display = XOpenDisplay( NULL ) ) )
	{
	  fprintf( stderr, "Cannot open display\n" );
	  exit(1);
	}

      screen = DefaultScreen( display );
  
      /* set windows attributes */
      set_win_attr.border_pixel = WhitePixel( display, screen );
      set_win_attr.background_pixel = BlackPixel( display, screen );
      set_win_attr.override_redirect = 0;
      set_win_attr.backing_store = WhenMapped;
      win_attr_mask = CWBorderPixel | CWBackPixel |
	CWOverrideRedirect | CWBackingStore;

      window = XCreateWindow( display, 
			      RootWindow( display, screen),
			      0, 0, Width*2, Height*2, 
			      0, DefaultDepth( display, screen ),
			      InputOutput,
			      DefaultVisual( display, screen ),
			      win_attr_mask,
			      &set_win_attr );

      /*
       * Set size hints & Properties
       */
      the_size_hints.width = Width*2; 
      the_size_hints.height = Height*2;
      the_size_hints.max_width = Width*2;
      the_size_hints.max_height = Height*2;
      the_size_hints.min_width = Width*2;
      the_size_hints.min_height = Height*2;
      the_size_hints.flags = PSize | PMaxSize | PMinSize;
      XSetStandardProperties( display, window,
			      "Turtlegraphics", "Turtlegraphics",
			      None, 0, 0,  &the_size_hints );
      
      /* SelectInput( display, window, ExposureMask | StructureNotifyMask ); */

      for (i=0; i<NUMBER(ColorNames); i++)
	if (ColorNames[i])
	  {
	    XColor	color,color1;
	    if (!XAllocNamedColor(display,
				  DefaultColormap(display,
						  DefaultScreen(display)),
				  ColorNames[i], &color, &color1))
	      {
		fprintf(stderr,"Unable to allocate color %s\n", 
			ColorNames[i]);
		if (strcmp(ColorNames[i],"black")==0)
		  GcValues.foreground=BlackPixel(display,
						 DefaultScreen(display));
		else
		  GcValues.foreground=WhitePixel(display,
						 DefaultScreen(display));
	      }
	    else
	      GcValues.foreground = color.pixel;
	    ColorGCs[i]=XCreateGC( display, window, GCForeground , &GcValues );
	  }
	else
      ColorGCs[i]=0;
    }
  XMapWindow(display, window);
  CurrentViewPort.xMin=0;
  CurrentViewPort.xMax=279;
  CurrentViewPort.yMin=0;
  CurrentViewPort.yMax=191;
  FillScreen(PENCOLOR_BLACK);
  CurrentTurtleX=279/2;
  CurrentTurtleY=191/2;
  CurrentTurtleAng=0;
  CurrentCharType=10;
  CurrentPenColor=PENCOLOR_NONE;
}

void TurtleGraphics(word EntryPoint)
{
  int	i;
  char	func[64];
  char	*p;

  i=0;
  for (i=0; i<sizeof(func); i++)
    if (!(func[i]=MemRdByte(EntryPoint,i)))
      break;

  p=&func[16];
  if (strcmp(p, "INITTURTLE")==0)
    InitTurtle(280,192);
  else if (strcmp(p, "TURN")==0)
    TurnTo(CurrentTurtleAng+PopInteger());
  else if (strcmp(p, "TURNTO")==0)
    TurnTo(PopInteger());
  else if (strcmp(p, "MOVE")==0)
    {
      Integer Dist=PopInteger();
      MoveTo(CurrentTurtleX+rint(cos(CurrentTurtleAng*3.14/180)*Dist),
	     CurrentTurtleY+rint(sin(CurrentTurtleAng*3.14/180)*Dist));
    }
  else if (strcmp(p, "MOVETO")==0)
    {
      Integer y=PopInteger();
      Integer x=PopInteger();
      MoveTo(x, y);
    }
  else if (strcmp(p, "PENCOLOR")==0)
    CurrentPenColor=Pop();
  else if (strcmp(p, "TEXTMODE")==0)
    {
      if (display)
	XUnmapWindow(display, window);
    }
  else if (strcmp(p, "GRAFMODE")==0)
    {
      if (display)
	XMapWindow(display, window);
    }
  else if (strcmp(p, "FILLSCREEN")==0)
    {
      if (display)
	FillScreen(Pop());
    }
  else if (strcmp(p, "VIEWPORT")==0)
    {
      CurrentViewPort.yMax=PopInteger();
      CurrentViewPort.yMin=PopInteger();
      CurrentViewPort.xMax=PopInteger();
      CurrentViewPort.xMin=PopInteger();
    }
  else if (strcmp(p, "TURTLEX")==0)
    {
      Pop();
      Pop();
      Push(CurrentTurtleX);
    }
  else if (strcmp(p, "TURTLEY")==0)
    {
      Pop();
      Pop();
      Push(CurrentTurtleY);
    }
  else if (strcmp(p, "TURTLEANG")==0)
    {
      Pop();
      Pop();
      Push(CurrentTurtleAng);
    }
  else if (strcmp(p, "SCREENBIT")==0)
    {
      Integer y=PopInteger();
      Integer x=PopInteger();
      if (display && (GetPixel(x,y)!=PENCOLOR_BLACK) )
	Push(1);
      else
	Push(0);
    }
  else if (strcmp(p, "DRAWBLOCK")==0)
    {
      int	x;
      int	y;
      Integer Mode    = PopInteger();
      Integer YScreen = PopInteger();
      Integer XScreen = PopInteger();
      Integer Height  = PopInteger();
      Integer Width   = PopInteger();
      Integer YSkip   = PopInteger();
      Integer XSkip   = PopInteger();
      Integer RowSize = PopInteger();
      word Source     = Pop();

      if (display)
	for (y=0; y<Height; y++)
	  {
	    int  Skip=XSkip&7;
	    for (x=0;x<Width;)
	      {
		int  w=Width-x;
		char Data = MemRdByte(Source, (y+YSkip)*RowSize+(x+XSkip)/8);

		if ( w > 8-Skip)
		  w = 8-Skip;

		DrawByte(Data, Skip, w, XScreen+x, YScreen+y, Mode);
		x+=w;
		Skip=0;
	      }
	  }
    }
  else if (strcmp(p, "WCHAR")==0)
    {
      char ch=Pop()&0xff;
      if (display)
	WChar(ch);
    }
  else if (strcmp(p, "WSTRING")==0)
    {
      word Addr=Pop();
      int  Len=MemRdByte(Addr, 0);
      int  i;

      if (display)
	for (i=0;i<Len; i++)
	  WChar(MemRdByte(Addr, i+1));
    }
  else if (strcmp(p, "CHARTYPE")==0)
    CurrentCharType=Pop()&0x0f;
  else
    panic("unknown turtlegraphic routine %s\n", p);

  if (display)
    XFlush(display);
}





